package tv.miniweb.gradle.plugins.dependencies;

import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.ivy.Ivy;
import org.apache.ivy.core.cache.ResolutionCacheManager;
import org.apache.ivy.core.module.id.ModuleId;
import org.apache.ivy.core.resolve.ResolveOptions;
import org.dom4j.Document;
import org.dom4j.Node;
import org.dom4j.io.SAXReader;
import org.gradle.api.Project;

public class DependencyTools
{
	/**
	 * Unfortunately, Ivy does not expose a proper interface for iterating over dependencies
	 * in anything other than a flat collection. To get a direct graph of depepndencies,
	 * we have to import the XML Ivy dependency report and run some xpath.
	 * 
	 * @param project
	 * @return List of directed graph root nodes
	 * @throws Exception
	 */
	@SuppressWarnings("unchecked")
	public static List<IvyDependency> generateDependencyTree(Project project, Ivy ivy) throws Exception
	{
		Map<String, IvyDependency> dependencyCache = new HashMap<String, IvyDependency>();
	
		ResolutionCacheManager cacheMgr = ivy.getResolutionCacheManager();
		
		String resolveId = ResolveOptions.getDefaultResolveId(new ModuleId((String) project.property("group"), project.getName()));
		
		File[] reports = cacheMgr.getConfigurationResolveReportsInCache(resolveId);
	
		List<IvyDependency> roots = new ArrayList<IvyDependency>();
		
		for(File report : reports)
		{
			Document doc = new SAXReader().read(report);
			
			List<Node> moduleNodes = doc.selectNodes("//module");
			
			String name, organisation, revision;
			
			Node infoNode = doc.selectSingleNode("//info");
			
			name = infoNode.selectSingleNode("@module").getText();
			organisation = infoNode.selectSingleNode("@organisation").getText();
			revision = infoNode.selectSingleNode("@revision").getText();
			
			IvyDependency root = getDependency(name, organisation, revision, dependencyCache);
			
			roots.add(root);
			
			for(Node module : moduleNodes)
			{
				name = module.selectSingleNode("@name").getText();
				organisation = module.selectSingleNode("@organisation").getText();
				
				List<Node> revisions = module.selectNodes("revision");

				for(Node revisionNode : revisions)
				{
					revision = module.selectSingleNode("revision/@name").getText();
					
					IvyDependency dependency = getDependency(name, organisation, revision, dependencyCache);
					
					List<Node> callers = revisionNode.selectNodes("caller");

					for(Node callerNode : callers)
					{
						name = callerNode.selectSingleNode("@name").getText();
						organisation = callerNode.selectSingleNode("@organisation").getText();
						revision = callerNode.selectSingleNode("@callerrev").getText();
						
						IvyDependency caller = getDependency(name, organisation, revision, dependencyCache);
						
						caller.addDependency(dependency);	
					}					
				}
			}
		}
		
		return roots;
	}
	
	/**
	 * Get an IvyDependency from the cach. If no such IvyDependency exists, create a new one
	 * and return that instead.
	 * 
	 * @param name
	 * @param organisation
	 * @param revision
	 * @param cache
	 * @return
	 */
	private static IvyDependency getDependency(String name, String organisation, String revision, Map<String, IvyDependency> cache)
	{
		String id = organisation + ":" + name + ":" + revision;
		
		IvyDependency dep = cache.get(id);
		
		if(dep == null)
		{
			dep = new IvyDependency(name, organisation, revision);
			
			cache.put(id, dep);
		}	
		
		return dep;
	}
}
