package tv.miniweb.gradle.plugins.dependencies;

import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.lang.reflect.Method;
import java.util.HashSet;
import java.util.Set;

import org.apache.ivy.Ivy;
import org.gradle.api.GradleException;
import org.gradle.api.Project;
import org.gradle.api.internal.ConventionTask;
import org.gradle.api.internal.dependencies.BaseDependencyManager;

/**
 * Produces dependency trees. Accepts two parameters:
 * 
 * -DoutputFormat={ascii|graphviz} chooses which format to produce a report in. Optional, defaults to ascii.
 * -DoutputFile=/path/to/file file to output to. Optional, defaults to std out
 * 
 * @author phil
 *
 */
public class ShowDependenciesTask extends ConventionTask
{

	@Override
	public void execute()
	{
		super.execute();
		
		OutputStream out;
		
		if(System.getProperty("outputFile") != null)
		{
			try
			{
				out = new FileOutputStream(System.getProperty("outputFile"));
			}
			catch (FileNotFoundException e)
			{
				throw new GradleException("Error opening output file for dependency report");
			}
		}
		else
		{
			out = System.out;
		}
		
		Writer writer = new OutputStreamWriter(out);
		
		DependencyReportOutputFormat format = DependencyReportOutputFormat.ASCII;
		
		if("graphviz".equals(System.getProperty("outputFormat")))
		{
			format = DependencyReportOutputFormat.GRAPHVIZ;
		}
		
		System.out.println("Generating " + format + " format report");

		for(Project project : getProject().getAllprojects())
		{
			System.out.println("Working on " + project.getName());
			try
			{

				if(format == DependencyReportOutputFormat.GRAPHVIZ)
				{
					renderDot(project, writer);
				}
				else
				{
					renderASCII(project, writer);
				}
			}
			catch (Exception e)
			{
				throw new GradleException("Error in building dependency graph." + e);
			}
		}
		
		try
		{
			writer.flush();
			out.flush();
			out.close();
		}
		catch(Exception e)
		{
			throw new GradleException("Error flushing and closing output");
		}
	}

	private void renderASCII(Project project, Writer writer) throws Exception
	{
		writer.write("###########################\n");
		writer.write(project.getName());
		writer.write("\n");
		
		for(IvyDependency root : DependencyTools.generateDependencyTree(project, getIvy(project)))
		{
			printDependencyTree(root, 0, writer);
		}
		
		writer.write("###########################\n");
	}

	
	
	/**
	 * Print a dependency tree in ASCII form
	 * 
	 * @param root
	 * @param depth
	 * @param writer
	 * @throws IOException
	 */
	private void printDependencyTree(IvyDependency root, int depth, Writer writer) throws IOException
	{
		writer.write(getIndent(depth));
		writer.write(root.toString());
		writer.write("\n");
		
		for(IvyDependency dep : root.getDependencies())
		{
			printDependencyTree(dep, depth + 1, writer);
		}
	}
	
	/**
	 * Render a dependency tree to Graphviz format
	 * 
	 * @param project
	 * @param writer
	 * @throws Exception
	 */
	private void renderDot(Project project, Writer writer) throws Exception
	{
		writer.write("digraph " + project.getName() + " {\n");
	
		Set<String> edges = new HashSet<String>();
		
		for(IvyDependency root : DependencyTools.generateDependencyTree(project, getIvy(project)))
		{
			buildDotDependencyTree(root, edges);
		}
		
		for(String edge : edges)
		{
			writer.write(edge + "\n");
		}
		
		writer.write("}\n" );
	}
	
	/**
	 * 
	 * @param root
	 * @param edges
	 */
	private void buildDotDependencyTree(IvyDependency root, Set<String> edges)
	{
		for(IvyDependency dep : root.getDependencies())
		{
			String edge = "\"" + root.getName() + "\" -> \"" + dep.getName().replace('-', '_') + "\";";
			edges.add(edge);
		}
		
		for(IvyDependency dep : root.getDependencies())
		{
			buildDotDependencyTree(dep, edges);
		}
	}
	
	/**
	 * Generate an approriate tab-indented string for the supplied depth
	 * 
	 * @param depth
	 * @return
	 */
	private String getIndent(int depth)
	{
		StringBuilder buffer = new StringBuilder();
		
		for(int x = 0; x < depth - 1; x++)
		{
			buffer.append("|\t");
		}
	
		buffer.append("|-----");
		
		return buffer.toString();
	}
	
	
	
	/**
	 * Hack method to get an Ivy instance
	 * 
	 * @return
	 */
	private Ivy getIvy(Project project)
	{
		try
		{
			BaseDependencyManager depManager = (BaseDependencyManager) project.getDependencies();
			
			depManager.resolve("runtime", false, true);
			
			Method getter = BaseDependencyManager.class.getDeclaredMethod("getIvy", null);
			getter.setAccessible(true);
			
			return (Ivy) getter.invoke(depManager, null);
		}
		catch(Exception e)
		{
			e.printStackTrace();
		}
		
		return null;
	}
	
	public ShowDependenciesTask(Project project, String name)
	{
		super(project, name);
	}
}
